import time

from django.core.management.base import BaseCommand

from dj_anonymizer.anonymizer import Anonymizer


class Command(BaseCommand):
    help = "Anonymize database according to provided patterns."

    def add_arguments(self, parser):
        parser.add_argument('-a', '--action',
                            help='"anonymize" models or "clean" models '
                                 '(default: both actions will be applied)')
        parser.add_argument('-c', '--check_only', action='store_true',
                            help='Check that all models are registered '
                                 'correctly')
        parser.add_argument('-o', '--only',
                            help='Execute specified for dj_anonymizer action '
                                 'only on specified model')
        parser.add_argument('-s', '--soft_mode', action='store_true',
                            help='Exception will not be raised if not all '
                                 'project models are registered')

    def handle(self, *args, **options):
        self.action = options['action']
        self.check_only = options['check_only']
        self.only = options['only']
        self.soft_mode = options['soft_mode']
        start = time.time()
        anonymizer = Anonymizer(
            soft_mode=self.soft_mode,
            only=self.only
        )
        self.stdout.write('Check pass successfully')
        self.stdout.write(f'{"=-" * 25}=')
        if self.check_only is True:
            return
        if self.action is None or self.action == 'anonymize':
            if (
                (self.only is None)
                or (self.only is not None
                    and self.only in Anonymizer.anonym_models)
            ) and len(anonymizer.anonym_models) > 0:
                self.stdout.write('Anonymizing started')
                anonymizer.anonymize(only=self.only)
                self.stdout.write('Anonymizing finished')
        if self.action is None or self.action == 'clean':
            if (
                (self.only is None)
                or (self.only is not None
                    and self.only in Anonymizer.clean_models)
            ) and len(anonymizer.clean_models) > 0:
                self.stdout.write('Cleaning started')
                anonymizer.clean(only=self.only)
                self.stdout.write('Cleaning finished')
        end = time.time() - start
        self.stdout.write(f'{"=-" * 25}=')
        self.stdout.write(f'Total time (sec.): {end:.2f}')



import django
from django.conf import settings

from dj_anonymizer.utils import import_if_exist, truncate_table


class Anonymizer:
    anonym_models = {}
    clean_models = {}
    skip_models = []

    def __init__(self, soft_mode=True, only=None):
        models_set = set()

        # this for django.contrib.*.models or can be used
        # as single file for defining models for anonymizing
        import_if_exist('base')

        for app in django.apps.apps.get_app_configs():
            models_set.update(
                Anonymizer.key(model)
                for model in app.get_models()
            )
            import_if_exist(app.name)

        all_models = set(
            self.skip_models +
            list(self.anonym_models.keys()) +
            list(self.clean_models.keys())
        )

        if only is not None and only not in all_models:
            raise LookupError(
                f'Model specified at --only attibute({only}) can\'t '
                f'be found at list of registered models'
            )

        if not soft_mode and not models_set.issubset(all_models):
            raise LookupError(
                f'You did not set those models to any list: '
                f'{list(models_set.difference(all_models))}'
            )

    @staticmethod
    def key(model):
        """
        Keys that used for referencing Django models in Anonymizer
        """
        return f'{model.__module__}.{model.__name__}'

    def anonymize(self, only=None):
        anon_list = self.anonym_models.values() if only is None \
            else [self.anonym_models[only]]
        # Size of chunks what will be used to select data from table.
        select_batch_size = getattr(
            settings, "ANONYMIZER_SELECT_BATCH_SIZE", 20000
        )
        # Size of chunks what will be used to update data in table.
        update_batch_size = getattr(
            settings, "ANONYMIZER_UPDATE_BATCH_SIZE", 500
        )
        for anonym_cls in anon_list:

            if not anonym_cls.get_fields_names():
                continue

            queryset = anonym_cls.Meta.queryset.only(
                *anonym_cls.get_fields_names()
            )
            print(f'Anonymizing model {self.key(queryset.model)}')

            i = 0
            total = queryset.count()
            for j in list(range(0, total,
                                select_batch_size)) + [None]:
                subset = queryset.order_by('pk')[i:j]
                for obj in subset:
                    i += 1

                    for name in anonym_cls.get_fields_names():
                        setattr(obj, name, next(
                            getattr(anonym_cls, name))
                        )
                queryset.bulk_update(
                    subset,
                    anonym_cls.get_fields_names(),
                    batch_size=update_batch_size,
                )

    def clean(self, only=None):
        clean_list = self.clean_models.values() if only is None \
            else [self.clean_models[only]]
        for queryset in clean_list:
            print(f'Cleaning {self.key(queryset.model)}')
            if getattr(queryset, 'truncate') is True:
                if getattr(queryset, 'cascade') is True:
                    truncate_table(queryset.model, True)
                else:
                    truncate_table(queryset.model, False)
            else:
                queryset.delete()



import importlib
import os

from django.conf import settings
from django.db import connections, router


VENDOR_TO_TRUNCATE = {
    'postgresql': 'TRUNCATE TABLE',
    'mysql': 'TRUNCATE TABLE',
    'sqlite': 'DELETE FROM',
    'oracle': 'TRUNCATE TABLE',
}

VENDOR_TO_CASCADE = {
    'postgresql': 'CASCADE',
    'oracle': 'CASCADE',
}


def import_if_exist(filename):
    """
    Check if file exist in appropriate path and import it
    """
    model_devinition_dir = getattr(
        settings,
        'ANONYMIZER_MODEL_DEFINITION_DIR',
        'anonymizer'
    )
    filepath = os.path.join(model_devinition_dir, filename)
    full_filepath = os.path.abspath(filepath + '.py')

    if os.path.isfile(full_filepath):
        spec = importlib.util.spec_from_file_location(filename, full_filepath)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)


def truncate_table(model, cascade=False):
    """
    Generate and execute via Django ORM proper SQL to truncate table
    """
    db = router.db_for_write(model)
    connection = connections[db]
    vendor = connection.vendor

    try:
        operation = VENDOR_TO_TRUNCATE[vendor]
    except KeyError:
        raise NotImplementedError(
            "Database vendor %s is not supported" % vendor
        )

    cascade_op = ''
    try:
        if cascade:
            cascade_op = VENDOR_TO_CASCADE[vendor]
    except KeyError:
        raise NotImplementedError(
            "DB vendor %s does not support TRUNCATE with CASCADE" % vendor
        )

    dbtable = '"{}"'.format(model._meta.db_table)

    sql = '{operation} {dbtable} {cascade}'.format(
        operation=operation,
        dbtable=dbtable,
        cascade=cascade_op,
    ).strip()
    with connection.cursor() as c:
        c.execute(sql)



import inspect

from django.db import ProgrammingError, connection
from django.db.models.fields import Field
from django.db.models.fields.related import (
    ForeignKey,
    ManyToManyField,
    OneToOneField
)

from dj_anonymizer.anonymizer import Anonymizer


class AnonymBase:
    truncate = False
    cascade = False

    def __init__(self, truncate=False, cascade=False):
        self.truncate = truncate
        self.cascade = cascade

    @classmethod
    def get_fields_names(cls):
        return [
            attr_name for attr_name in dir(cls)
            if inspect.isgenerator(getattr(cls, attr_name))
        ]

    @classmethod
    def get_relation_fields(cls, model):
        relation_fields = [field.name for field in model._meta.get_fields()
                           if isinstance(field, (ManyToManyField,
                                                 OneToOneField,
                                                 ForeignKey))
                           ]
        relation_fields.append(model._meta.pk.name)
        return relation_fields

    @classmethod
    def init_meta(cls, model):
        if hasattr(cls.Meta, 'queryset'):
            if cls.Meta.queryset.model != model:
                raise TypeError(
                    f'Class {Anonymizer.key(cls.Meta.queryset.model)} does '
                    f'not belong to the allowed {Anonymizer.key(AnonymBase)}'
                )
        else:
            setattr(cls.Meta, 'queryset', model.objects.all())


def register_anonym(models):
    for model, cls_anonym in models:

        if model._meta.db_table not in connection.introspection.table_names():
            raise ProgrammingError(
                f'Model {Anonymizer.key(model)} '
                f'registered, but table does not exist'
            )

        cls_anonym.init_meta(model)

        anonym_fields = set(cls_anonym.get_fields_names())
        model_fields = set(
            field.name for field in model._meta.get_fields()
            if isinstance(field, Field)
        )

        if hasattr(cls_anonym.Meta, 'exclude_fields'):
            exclude_fields = set(cls_anonym.Meta.exclude_fields)
        else:
            exclude_fields = model_fields - anonym_fields

        exclude_fields.update(cls_anonym.get_relation_fields(model))

        if exclude_fields & anonym_fields:
            raise LookupError(
                f'Fields {list(exclude_fields & anonym_fields)} of model '
                f'{Anonymizer.key(model)} are present in both anonymization '
                f'and excluded lists'
            )

        specified_fields = exclude_fields | anonym_fields

        if specified_fields < model_fields:
            raise LookupError(
                f'Fields {list(model_fields - specified_fields)} were not '
                f'registered in {Anonymizer.key(cls_anonym)} class for '
                f'{Anonymizer.key(model)} model'
            )
        if specified_fields > model_fields:
            raise LookupError(
                f'Fields {list(specified_fields - model_fields)} present in '
                f'{Anonymizer.key(cls_anonym)} class, but does not exist in '
                f'{Anonymizer.key(model)} model'''
            )
        if specified_fields != model_fields:
            raise LookupError(
                f'Fields in {Anonymizer.key(cls_anonym)} are not '
                f'the same as in {Anonymizer.key(model)} Check spelling'
            )

        if Anonymizer.key(model) in Anonymizer.anonym_models.keys():
            raise ValueError(
                f'Model {Anonymizer.key(model)} '
                f'is already declared in register_anonym'
            )
        Anonymizer.anonym_models[Anonymizer.key(model)] = cls_anonym


def register_clean(models):
    for model, cls_anonym in models:
        if not (cls_anonym == AnonymBase
                or isinstance(cls_anonym, AnonymBase)):
            raise TypeError(
                f'Class used for cleaning model {Anonymizer.key(model)} does '
                f'not belong to the allowed {Anonymizer.key(AnonymBase)}'
            )

        if model._meta.db_table not in connection.introspection.table_names():
            raise ProgrammingError(
                f'Model {Anonymizer.key(model)} '
                f'registered, but table does not exist'
            )

        queryset = model.objects.all()
        queryset.truncate = cls_anonym.truncate
        queryset.cascade = cls_anonym.cascade
        if Anonymizer.key(model) in Anonymizer.clean_models.keys():
            raise ValueError(
                f'Model {Anonymizer.key(model)} '
                f'is already declared in register_clean'
            )
        Anonymizer.clean_models[Anonymizer.key(model)] = queryset


def register_skip(models):
    for model in models:
        if Anonymizer.key(model) in Anonymizer.skip_models:
            raise ValueError(
                f'Model {Anonymizer.key(model)} '
                f'is already declared in register_skip'
            )
        Anonymizer.skip_models.append(Anonymizer.key(model))



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import os


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name='dj_anonymizer',
    packages=['dj_anonymizer'],
    include_package_data=True,
    version='0.6.1',
    description='dj_anonymizer is a utility designed to anonymize '
                + 'production databases with various types of mock data, '
                + 'specifically designed for use within Django projects.',
    long_description=(read('README.md')),
    long_description_content_type='text/markdown',
    license='MIT',
    author='Tim Pagurets',
    url='https://github.com/preply/dj_anonymizer',
    keywords=['django', 'data', 'database', 'anonymization', 'dj-anonymizer'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Framework :: Django',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Topic :: Database'
    ],
    install_requires=[
        'Django>=4.2',
    ]
)



import itertools

from django.contrib.auth.hashers import make_password
from django.template.defaultfilters import slugify


def function(callback, *args, **kwargs):
    for _ in itertools.count():
        yield callback(*args, **kwargs)


def password(password, *args, **kwargs):
    for _ in itertools.count():
        yield make_password(password, *args, **kwargs)


def string(field_value, seq_start=0, seq_step=1, seq_callback=None,
           seq_args=(), seq_kwargs=None, seq_slugify=True):
    if seq_kwargs is None:
        seq_kwargs = {}

    for i in itertools.count(seq_start, seq_step):
        if seq_callback:
            seq = seq_callback(*seq_args, **seq_kwargs)
            if seq_slugify:
                seq = slugify(seq)
        else:
            seq = i

        yield field_value.format(seq=seq)
